# SPDX-License-Identifier: LGPL-3.0-or-later
"""DP-GUI entrypoint."""


def start_dpgui(*, port: int, bind_all: bool, **kwargs):
    """Host DP-GUI server.

    Parameters
    ----------
    port : int
        The port to serve DP-GUI on.
    bind_all : bool
        Serve on all public interfaces. This will expose your DP-GUI instance
        to the network on both IPv4 and IPv6 (where available).
    **kwargs
        additional arguments

    Raises
    ------
    ModuleNotFoundError
        The dpgui package is not installed
    """
    try:
        from dpgui import (
            start_dpgui,
        )
    except ModuleNotFoundError as e:
        raise ModuleNotFoundError(
            "To use DP-GUI, please install the dpgui package:\npip install dpgui"
        ) from e
    start_dpgui(port=port, bind_all=bind_all)



from enum import IntEnum


class JobStatus(IntEnum):
    unsubmitted = 1
    waiting = 2
    running = 3
    terminated = 4
    finished = 5
    completing = 6
    unknown = 100


#     def __str__(self):
#         return repr(self)



import os
import sys
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
# from .context import LocalSession
# from .context import LocalContext
from .context import (
    Job,
    Submission,
    setUpModule,  # noqa: F401
)
from .sample_class import SampleClass


class TestJob(unittest.TestCase):
    def setUp(self):
        self.job = SampleClass.get_sample_job()

        self.submission2 = Submission.submission_from_json("jsons/submission.json")
        self.job2 = self.submission2.belonging_jobs[0]

    def test_eq(self):
        self.assertTrue(self.job == self.job2)

    def test_get_hash(self):
        self.assertEqual(self.job.get_hash(), self.job2.get_hash())
        # self.assertEqual(self.submission, self.submission2)

    def test_serialize_deserialize(self):
        self.assertEqual(self.job, Job.deserialize(job_dict=self.job.serialize()))

    def test_static_serialize(self):
        self.assertNotIn(
            "job_state", list(self.job.serialize(if_static=True).values())[0]
        )
        self.assertNotIn("job_id", list(self.job.serialize(if_static=True).values())[0])
        self.assertNotIn(
            "fail_count", list(self.job.serialize(if_static=True).values())[0]
        )

    def test_get_job_state(self):
        pass

    def test_handle_unexpected_job_state(self):
        pass

    def test_register_job_id(self):
        pass

    def test_submit_job(self):
        pass

    def test_job_to_json(self):
        pass


#  def test_content_serialize(self):
#      self.assertEqual(self.job.content_serialize(), self.job.serialize()[self.job.job_hash])



import os
import shutil
import uuid

import tqdm

try:
    from bohriumsdk.client import Client
    from bohriumsdk.job import Job
    from bohriumsdk.storage import Storage
    from bohriumsdk.util import Util
except ModuleNotFoundError:
    found_bohriumsdk = False
else:
    found_bohriumsdk = True

from dpdispatcher import dlog
from dpdispatcher.base_context import BaseContext
from dpdispatcher.JobStatus import JobStatus

DP_CLOUD_SERVER_HOME_DIR = os.path.join(
    os.path.expanduser("~"), ".dpdispatcher/", "dp_cloud_server/"
)


class OpenAPIContext(BaseContext):
    def __init__(
        self,
        local_root,
        remote_root=None,
        remote_profile={},
        *args,
        **kwargs,
    ):
        if not found_bohriumsdk:
            raise ModuleNotFoundError(
                "bohriumsdk not installed. Install dpdispatcher with `pip install dpdispatcher[bohrium]`"
            )
        self.init_local_root = local_root
        self.init_remote_root = remote_root
        self.temp_local_root = os.path.abspath(local_root)
        self.remote_profile = remote_profile
        self.client = Client()
        self.storage = Storage(client=self.client)
        self.job = Job(client=self.client)
        self.util = Util()
        self.jgid = None

    @classmethod
    def load_from_dict(cls, context_dict):
        local_root = context_dict.get("local_root", "./")
        remote_root = context_dict.get("remote_root", None)
        remote_profile = context_dict.get("remote_profile", {})

        bohrium_context = cls(
            local_root=local_root,
            remote_root=remote_root,
            remote_profile=remote_profile,
        )
        return bohrium_context

    def bind_submission(self, submission):
        self.submission = submission
        self.local_root = os.path.join(self.temp_local_root, submission.work_base)
        self.remote_root = "."

        self.submission_hash = submission.submission_hash

        self.machine = submission.machine

    def _gen_object_key(self, job, zip_filename):
        if hasattr(job, "upload_path") and job.upload_path:
            return job.upload_path
        else:
            project_id = self.remote_profile.get("project_id")

            uid = uuid.uuid4()
            path = os.path.join(str(project_id), str(uid), zip_filename)
            setattr(job, "upload_path", path)
            return path

    def upload_job(self, job, common_files=None):
        if common_files is None:
            common_files = []
        self.machine.gen_local_script(job)
        zip_filename = job.job_hash + ".zip"
        zip_task_file = os.path.join(self.local_root, zip_filename)

        upload_file_list = [
            job.script_file_name,
            f"{job.script_file_name}.run",
        ]

        upload_file_list.extend(common_files)

        for task in job.job_task_list:
            for file in task.forward_files:
                upload_file_list.append(os.path.join(task.task_work_path, file))

        upload_zip = Util.zip_file_list(
            self.local_root, zip_task_file, file_list=upload_file_list
        )
        project_id = self.remote_profile.get("project_id", 0)

        data = self.job.create(
            project_id=project_id,
            name=self.remote_profile.get("job_name", "DP-GEN"),
            group_id=self.jgid,  # type: ignore
        )
        self.jgid = data.get("jobGroupId", "")  # type: ignore
        token = data.get("token", "")  # type: ignore

        object_key = os.path.join(data["storePath"], zip_filename)  # type: ignore
        job.upload_path = object_key
        job.job_id = data["jobId"]  # type: ignore
        job.jgid = data["jobGroupId"]  # type: ignore
        self.storage.upload_From_file_multi_part(
            object_key=object_key, file_path=upload_zip, token=token
        )

        # self._backup(self.local_root, upload_zip)

    def upload(self, submission):
        # oss_task_dir = os.path.join('%s/%s/%s.zip' % ('indicate', file_uuid, file_uuid))
        # zip_filename = submission.submission_hash + '.zip'
        # oss_task_zip = 'indicate/' + submission.submission_hash + '/' + zip_filename

        # zip_path = "/home/felix/workplace/22_dpdispatcher/dpdispatcher-yfb/dpdispatcher/dpcloudserver/t.txt"
        # zip_path = self.local_root
        bar_format = "{l_bar}{bar}| {n:.02f}/{total:.02f} %  [{elapsed}<{remaining}, {rate_fmt}{postfix}]"
        job_to_be_uploaded = []
        result = None
        dlog.info("checking all job has been uploaded")
        for job in submission.belonging_jobs:
            if job.job_state in (JobStatus.unsubmitted, JobStatus.terminated):
                job_to_be_uploaded.append(job)
        if len(job_to_be_uploaded) == 0:
            dlog.info("all job has been uploaded, continue")
            return result
        for job in tqdm.tqdm(
            job_to_be_uploaded,
            desc="Uploading to tiefblue",
            bar_format=bar_format,
            leave=False,
            disable=None,
        ):
            self.upload_job(job, submission.forward_common_files)
        return result
        # return oss_task_zip
        # api.upload(self.oss_task_dir, zip_task_file)

    def download(self, submission):
        jobs = submission.belonging_jobs
        job_hashs = {}
        job_infos = {}
        job_result = []
        for job in jobs:
            jid = job.job_id
            job_hashs[jid] = job.job_hash
            jobinfo = self.job.detail(jid)
            # jobinfo = self.api.get_job_detail(jid)
            job_result.append(jobinfo)
        # if group_id is not None:
        #     job_result = self.api.get_tasks_list(group_id)
        for each in job_result:
            if "resultUrl" in each and each["resultUrl"] != "" and each["status"] == 2:
                job_hash = ""
                if each["id"] not in job_hashs:
                    dlog.info(
                        f"find unexpect job_hash, but task {each['id']} still been download."
                    )
                    dlog.debug(str(job_hashs))
                    job_hash = str(each["id"])
                else:
                    job_hash = job_hashs[each["id"]]
                job_infos[job_hash] = each
        bar_format = "{l_bar}{bar}| {n:.02f}/{total:.02f} %  [{elapsed}<{remaining}, {rate_fmt}{postfix}]"
        for job_hash, info in tqdm.tqdm(
            job_infos.items(),
            desc="Validating download file from Lebesgue",
            bar_format=bar_format,
            leave=False,
            disable=None,
        ):
            result_filename = job_hash + "_back.zip"
            target_result_zip = os.path.join(self.local_root, result_filename)
            if self._check_if_job_has_already_downloaded(
                target_result_zip, self.local_root
            ):
                continue
            self.storage.download_from_url(info["resultUrl"], target_result_zip)
            Util.unzip_file(target_result_zip, out_dir=self.local_root)
            self._backup(self.local_root, target_result_zip)
        self._clean_backup(
            self.local_root, keep_backup=self.remote_profile.get("keep_backup", True)
        )
        return True

    def write_file(self, fname, write_str):
        result = self.write_home_file(fname, write_str)
        return result

    def write_local_file(self, fname, write_str):
        local_filename = os.path.join(self.local_root, fname)
        with open(local_filename, "w") as f:
            f.write(write_str)
        return local_filename

    def read_file(self, fname):
        result = self.read_home_file(fname)
        return result

    def write_home_file(self, fname, write_str):
        # os.makedirs(self.remote_root, exist_ok = True)
        with open(os.path.join(DP_CLOUD_SERVER_HOME_DIR, fname), "w") as fp:
            fp.write(write_str)
        return True

    def read_home_file(self, fname):
        with open(os.path.join(DP_CLOUD_SERVER_HOME_DIR, fname)) as fp:
            ret = fp.read()
        return ret

    def check_file_exists(self, fname):
        result = self.check_home_file_exits(fname)
        return result

    def check_home_file_exits(self, fname):
        return os.path.isfile(os.path.join(DP_CLOUD_SERVER_HOME_DIR, fname))

    def clean(self):
        submission_file_name = f"{self.submission.submission_hash}.json"
        submission_json = os.path.join(DP_CLOUD_SERVER_HOME_DIR, submission_file_name)
        os.remove(submission_json)
        return True

    def _check_if_job_has_already_downloaded(self, target, local_root):
        backup_file_location = os.path.join(
            local_root, "backup", os.path.split(target)[1]
        )
        if os.path.exists(backup_file_location):
            return True
        else:
            return False

    def _backup(self, local_root, target):
        try:
            # move to backup directory
            os.makedirs(os.path.join(local_root, "backup"), exist_ok=True)
            shutil.move(
                target, os.path.join(local_root, "backup", os.path.split(target)[1])
            )
        except (OSError, shutil.Error) as e:
            dlog.exception("unable to backup file, " + str(e))

    def _clean_backup(self, local_root, keep_backup=True):
        if not keep_backup:
            dir_to_be_removed = os.path.join(local_root, "backup")
            if os.path.exists(dir_to_be_removed):
                shutil.rmtree(dir_to_be_removed)



#!/usr/bin/env python
import argparse
from typing import List, Optional

from dpdispatcher.gui import start_dpgui


def main_parser() -> argparse.ArgumentParser:
    """Dpdispatcher commandline options argument parser.

    Notes
    -----
    This function is used by documentation.

    Returns
    -------
    argparse.ArgumentParser
        the argument parser
    """
    parser = argparse.ArgumentParser(
        description="dpdispatcher: Generate HPC scheduler systems jobs input scripts, submit these scripts to HPC systems, and poke until they finish",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = parser.add_subparsers(title="Valid subcommands", dest="command")
    ##########################################
    # gui
    parser_gui = subparsers.add_parser(
        "gui",
        help="Serve DP-GUI.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_gui.add_argument(
        "-p",
        "--port",
        type=int,
        default=6042,
        help="The port to serve DP-GUI on.",
    )
    parser_gui.add_argument(
        "--bind_all",
        action="store_true",
        help=(
            "Serve on all public interfaces. This will expose your DP-GUI instance "
            "to the network on both IPv4 and IPv6 (where available)."
        ),
    )
    return parser


def parse_args(args: Optional[List[str]] = None):
    """Dpdispatcher commandline options argument parsing.

    Parameters
    ----------
    args : List[str]
        list of command line arguments, main purpose is testing default option None
        takes arguments from sys.argv
    """
    parser = main_parser()

    parsed_args = parser.parse_args(args=args)
    if parsed_args.command is None:
        parser.print_help()

    return parsed_args


def main():
    args = parse_args()
    if args.command == "gui":
        start_dpgui(
            port=args.port,
            bind_all=args.bind_all,
        )
    elif args.command is None:
        pass
    else:
        raise RuntimeError(f"unknown command {args.command}")


if __name__ == "__main__":
    main()



import base64
import hashlib
import hmac
import os
import struct
import subprocess
import time
from typing import TYPE_CHECKING, Callable, Optional, Type, Union

from dpdispatcher import dlog

if TYPE_CHECKING:
    from dpdispatcher import Resources


def get_sha256(filename):
    """Get sha256 of a file.

    Parameters
    ----------
    filename : str
        The filename.

    Returns
    -------
    sha256: str
        The sha256.
    """
    h = hashlib.sha256()
    # buffer size: 128 kB
    b = bytearray(128 * 1024)
    mv = memoryview(b)
    with open(filename, "rb", buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            h.update(mv[:n])
    sha256 = h.hexdigest()
    return sha256


def hotp(key: str, period: int, token_length: int = 6, digest="sha1"):
    key_ = base64.b32decode(key.upper() + "=" * ((8 - len(key)) % 8))
    period_ = struct.pack(">Q", period)
    mac = hmac.new(key_, period_, digest).digest()
    offset = mac[-1] & 0x0F
    binary = struct.unpack(">L", mac[offset : offset + 4])[0] & 0x7FFFFFFF
    return str(binary)[-token_length:].zfill(token_length)


def generate_totp(secret: str, period: int = 30, token_length: int = 6) -> str:
    """Generate time-based one time password (TOTP) from the secret.

    Some HPCs use TOTP for two-factor authentication for safety.

    Parameters
    ----------
    secret : str
        The encoded secret provided by the HPC. It's usually extracted
        from a 2D code and base32 encoded.
    period : int, default=30
        Time period where the code is valid in seconds.
    token_length : int, default=6
        The token length.

    Returns
    -------
    token: str
        The generated token.

    References
    ----------
    https://github.com/lepture/otpauth/blob/49914d83d36dbcd33c9e26f65002b21ce09a6303/otpauth.py#L143-L160
    """
    digest = "sha1"
    return hotp(secret, int(time.time() / period), token_length, digest)


def run_cmd_with_all_output(cmd, shell=True):
    with subprocess.Popen(
        cmd, shell=shell, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as proc:
        out, err = proc.communicate()
        ret = proc.returncode
    return (ret, out, err)


def rsync(
    from_file: str,
    to_file: str,
    port: int = 22,
    key_filename: Optional[str] = None,
    timeout: Union[int, float] = 10,
):
    """Call rsync to transfer files.

    Parameters
    ----------
    from_file : str
        SRC
    to_file : str
        DEST
    port : int, default=22
        port for ssh
    key_filename : str, optional
        identity file name
    timeout : int, default=10
        timeout for ssh

    Raises
    ------
    RuntimeError
        when return code is not 0
    """
    ssh_cmd = [
        "ssh",
        "-o",
        "ConnectTimeout=" + str(timeout),
        "-o",
        "BatchMode=yes",
        "-o",
        "StrictHostKeyChecking=no",
        "-p",
        str(port),
        "-q",
    ]
    if key_filename is not None:
        ssh_cmd.extend(["-i", key_filename])
    cmd = [
        "rsync",
        # -a: archieve
        # -z: compress
        "-az",
        "-e",
        " ".join(ssh_cmd),
        "-q",
        from_file,
        to_file,
    ]
    ret, out, err = run_cmd_with_all_output(cmd, shell=False)
    if ret != 0:
        raise RuntimeError(f"Failed to run {cmd}: {err}")


class RetrySignal(Exception):
    """Exception to give a signal to retry the function."""


def retry(
    max_retry: int = 3,
    sleep: Union[int, float] = 60,
    catch_exception: Type[BaseException] = RetrySignal,
) -> Callable:
    """Retry the function until it succeeds or fails for certain times.

    Parameters
    ----------
    max_retry : int, default=3
        The maximum retry times. If None, it will retry forever.
    sleep : int or float, default=60
        The sleep time in seconds.
    catch_exception : Exception, default=Exception
        The exception to catch.

    Returns
    -------
    decorator: Callable
        The decorator.

    Examples
    --------
    >>> @retry(max_retry=3, sleep=60, catch_exception=RetrySignal)
    ... def func():
    ...     raise RetrySignal("Failed")
    """

    def decorator(func):
        assert max_retry > 0, "max_retry must be greater than 0"

        def wrapper(*args, **kwargs):
            current_retry = 0
            errors = []
            while max_retry is None or current_retry < max_retry:
                try:
                    return func(*args, **kwargs)
                except (catch_exception,) as e:
                    errors.append(e)
                    dlog.exception("Failed to run %s: %s", func.__name__, e)
                    # sleep certain seconds
                    dlog.warning("Sleep %s s and retry...", sleep)
                    time.sleep(sleep)
                    current_retry += 1
            else:
                # raise all exceptions
                raise RuntimeError(
                    "Failed to run %s for %d times" % (func.__name__, current_retry)
                ) from errors[-1]

        return wrapper

    return decorator


def customized_script_header_template(
    filename: os.PathLike, resources: "Resources"
) -> str:
    with open(filename) as f:
        template = f.read()
    return template.format(**resources.serialize())



import os
import shutil
import sys
import time
import unittest
import uuid

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "slurm"
from .context import (
    JobStatus,
    LocalContext,
    LocalSession,
    Slurm,
)


@unittest.skipIf(not shutil.which("sbatch"), "requires Slurm")
class TestSlurm(unittest.TestCase):
    def setUp(self):
        os.makedirs("loc", exist_ok=True)
        os.makedirs("rmt", exist_ok=True)
        os.makedirs("loc/task0", exist_ok=True)
        os.makedirs("loc/task1", exist_ok=True)
        for ii in ["loc/task0", "loc/task1"]:
            with open(os.path.join(ii, "test0"), "w") as fp:
                fp.write(str(uuid.uuid4()))
        work_profile = LocalSession({"work_path": "rmt"})
        self.ctx = LocalContext("loc", work_profile)
        self.slurm = Slurm(self.ctx)

    def tearDown(self):
        shutil.rmtree("loc")
        shutil.rmtree("rmt")
        if os.path.exists("dpgen.log"):
            os.remove("dpgen.log")

    def test_gen_sub_script(self):
        job_dirs = ["task0", "task1"]
        self.slurm.context.upload(job_dirs, ["test0"])
        ret = self.slurm.sub_script(job_dirs, ["touch test1", "touch test2"])
        self.slurm.context.write_file("run.sub", ret)
        with open("run.sub", "w") as fp:
            fp.write(ret)

    def test_sub_success(self):
        job_dirs = ["task0", "task1"]
        self.slurm.context.upload(job_dirs, ["test0"])
        self.slurm.submit(job_dirs, ["touch test1", "touch test2"])
        while True:
            ret = self.slurm.check_status()
            if ret == JobStatus.finished:
                break
            time.sleep(1)
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task0/tag_0_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task0/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task1/tag_0_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task1/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt",
                    self.slurm.context.remote_root,
                    "%s_tag_finished" % self.slurm.context.job_uuid,
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task0/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task1/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task0/test2")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task1/test2")
            )
        )

    def test_sub_scancel(self):
        job_dirs = ["task0", "task1"]
        self.slurm.context.upload(job_dirs, ["test0"])
        # sub
        self.slurm.submit(job_dirs, ["touch test1", "sleep 10"])
        while True:
            ret = self.slurm.check_status()
            if ret == JobStatus.finished:
                raise RuntimeError("should not finished")
            if ret == JobStatus.running:
                # wait for file writing
                time.sleep(2)
                job_id = self.slurm._get_job_id()
                os.system("scancel " + job_id)
                break
            time.sleep(1)
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task0/tag_0_finished"
                )
            )
        )
        self.assertFalse(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task0/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task1/tag_0_finished"
                )
            )
        )
        self.assertFalse(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task1/tag_1_finished"
                )
            )
        )
        self.assertFalse(
            os.path.isfile(
                os.path.join(
                    "rmt",
                    self.slurm.context.remote_root,
                    "%s_tag_finished" % self.slurm.context.job_uuid,
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task0/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task1/test1")
            )
        )
        self.assertFalse(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task0/test2")
            )
        )
        self.assertFalse(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task1/test2")
            )
        )
        # sub restart
        self.slurm.submit(job_dirs, ["rm test1", "touch test2"], restart=True)
        while True:
            ret = self.slurm.check_status()
            if ret == JobStatus.finished:
                break
            time.sleep(1)
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task0/tag_0_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task0/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task1/tag_0_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt", self.slurm.context.remote_root, "task1/tag_1_finished"
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    "rmt",
                    self.slurm.context.remote_root,
                    "%s_tag_finished" % self.slurm.context.job_uuid,
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task0/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task1/test1")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task0/test2")
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join("rmt", self.slurm.context.remote_root, "task1/test2")
            )
        )



import os
import shutil
import time
import uuid
import warnings

from dpdispatcher import dlog
from dpdispatcher.dpcloudserver import Client, zip_file
from dpdispatcher.dpcloudserver.config import ALI_OSS_BUCKET_URL
from dpdispatcher.JobStatus import JobStatus
from dpdispatcher.machine import Machine
from dpdispatcher.utils import customized_script_header_template

shell_script_header_template = """
#!/bin/bash -l
"""


class Bohrium(Machine):
    alias = ("Lebesgue", "DpCloudServer")

    def __init__(self, context):
        self.context = context
        self.input_data = context.remote_profile["input_data"].copy()
        self.api_version = 2
        if "api_version" in self.input_data:
            self.api_version = self.input_data.get("api_version", 2)
        if "lebesgue_version" in self.input_data:
            self.api_version = self.input_data.get("lebesgue_version", 2)
        self.grouped = self.input_data.get("grouped", False)
        email = context.remote_profile.get("email", None)
        phone = context.remote_profile.get("phone", None)
        username = context.remote_profile.get("username", None)
        password = context.remote_profile.get("password", None)
        self.retry_count = context.remote_profile.get("retry_count", 3)
        self.ignore_exit_code = context.remote_profile.get("ignore_exit_code", True)

        ticket = os.environ.get("BOHR_TICKET", None)
        if ticket:
            self.api = Client(ticket=ticket)
            self.group_id = None
            return

        if email is None and username is not None:
            raise DeprecationWarning(
                "username is no longer support in current version, "
                "please consider use email instead of username."
            )
        if email is None and phone is None:
            raise ValueError(
                "can not find email/phone number in remote_profile, please check your machine file."
            )

        if password is None:
            raise ValueError(
                "can not find password in remote_profile, please check your machine file."
            )
        if self.api_version == 1:
            raise DeprecationWarning(
                "api version 1 is deprecated. Use version 2 instead."
            )

        account = email
        if email is None:
            account = phone
        self.api = Client(account, password)

        self.group_id = None

    def gen_script(self, job):
        shell_script = super(DpCloudServer, self).gen_script(job)
        return shell_script

    def gen_script_header(self, job):
        resources = job.resources
        if (
            resources["strategy"].get("customized_script_header_template_file")
            is not None
        ):
            shell_script_header = customized_script_header_template(
                resources["strategy"]["customized_script_header_template_file"],
                resources,
            )
        else:
            shell_script_header = shell_script_header_template
        return shell_script_header

    def gen_local_script(self, job):
        script_str = self.gen_script(job)
        script_file_name = job.script_file_name
        self.context.write_local_file(fname=script_file_name, write_str=script_str)
        script_run_str = self.gen_script_command(job)
        script_run_file_name = f"{job.script_file_name}.run"
        self.context.write_local_file(
            fname=script_run_file_name, write_str=script_run_str
        )
        return script_file_name

    def _gen_backward_files_list(self, job):
        result_file_list = []
        # result_file_list.extend(job.backward_common_files)
        for task in job.job_task_list:
            result_file_list.extend(
                [os.path.join(task.task_work_path, b_f) for b_f in task.backward_files]
            )
        result_file_list = list(set(result_file_list))
        return result_file_list

    def _gen_oss_path(self, job, zip_filename):
        if hasattr(job, "upload_path") and job.upload_path:
            return job.upload_path
        else:
            program_id = self.context.remote_profile.get("program_id")
            program_id = self.context.remote_profile.get("project_id", program_id)
            if program_id is None:
                dlog.info(
                    "can not find program id in remote profile, upload to default program id."
                )
                program_id = 0
            uid = uuid.uuid4()
            path = os.path.join("program", str(program_id), str(uid), zip_filename)
            setattr(job, "upload_path", path)
            return path

    def do_submit(self, job):
        self.gen_local_script(job)
        zip_filename = job.job_hash + ".zip"
        # oss_task_zip = 'indicate/' + job.job_hash + '/' + zip_filename
        oss_task_zip = self._gen_oss_path(job, zip_filename)
        job_resources = ALI_OSS_BUCKET_URL + oss_task_zip
        input_data = self.input_data.copy()

        if not input_data.get("job_resources"):
            input_data["job_resources"] = []
        input_data["job_resources"].append(job_resources)
        input_data["command"] = f"bash {job.script_file_name}"
        if not input_data.get("backward_files"):
            input_data["backward_files"] = self._gen_backward_files_list(job)
        input_data["logFiles"] = os.path.join(
            job.job_task_list[0].task_work_path, job.job_task_list[0].outlog
        )
        program_id = self.context.remote_profile.get("program_id")
        program_id = self.context.remote_profile.get("project_id", program_id)
        if program_id is None:
            warnings.warn("program_id is compulsory.")
        job_id, group_id = self.api.job_create(
            job_type=input_data["job_type"],
            oss_path=input_data["job_resources"],
            input_data=input_data,
            program_id=program_id,
            group_id=self.group_id,
        )
        if self.grouped:
            self.group_id = group_id
        job.job_id = str(job_id) + ":job_group_id:" + str(group_id)
        job_id = job.job_id
        job.job_state = JobStatus.waiting
        return job_id

    def _get_job_detail(self, job_id, group_id):
        check_return = self.api.get_job_detail(job_id)
        assert check_return is not None, (
            f"Failed to retrieve tasks information. To resubmit this job, please "
            f"try again, if this problem still exists please delete the submission "
            f"file and try again.\nYou can check submission.submission_hash in the "
            f'previous log or type `grep -rl "{job_id}:job_group_id:{group_id}" '
            f"~/.dpdispatcher/dp_cloud_server/` to find corresponding file. "
            f"You can try with command:\n    "
            f'rm $(grep -rl "{job_id}:job_group_id:{group_id}" ~/.dpdispatcher/dp_cloud_server/)'
        )
        return check_return

    def check_status(self, job):
        if job.job_id == "":
            return JobStatus.unsubmitted
        job_id = job.job_id
        group_id = None
        if isinstance(job.job_id, str) and ":job_group_id:" in job.job_id:
            group_id = None
            ids = job.job_id.split(":job_group_id:")
            job_id, group_id = int(ids[0]), int(ids[1])
            if (
                self.input_data.get("grouped")
                and ":job_group_id:" not in self.input_data
            ):
                self.group_id = group_id
            self.api_version = 2
        dlog.debug(
            f"debug: check_status; job.job_id:{job_id}; job.job_hash:{job.job_hash}"
        )
        check_return = self._get_job_detail(job_id, group_id)
        try:
            dp_job_status = check_return["status"]
        except IndexError as e:
            dlog.error(
                f"cannot find job information in bohrium for job {job.job_id}. check_return:{check_return}; retry one more time after 60 seconds"
            )
            time.sleep(60)
            retry_return = self._get_job_detail(job_id, group_id)
            try:
                dp_job_status = retry_return["status"]
            except IndexError as e:
                raise RuntimeError(
                    f"cannot find job information in bohrium for job {job.job_id} {check_return} {retry_return}"
                )

        job_state = self.map_dp_job_state(
            dp_job_status, check_return.get("exitCode", 0), self.ignore_exit_code
        )
        if job_state == JobStatus.finished:
            job_log = self.api.get_log(job_id)
            if self.input_data.get("output_log"):
                print(job_log, end="")
            self._download_job(job)
        elif self.input_data.get("output_log") and job_state == JobStatus.running:
            job_log = self.api.get_log(job_id)
            print(job_log, end="")
        return job_state

    def _download_job(self, job):
        job_url = self.api.get_job_result_url(job.job_id)
        if not job_url:
            return
        job_hash = job.job_hash
        result_filename = job_hash + "_back.zip"
        target_result_zip = os.path.join(self.context.local_root, result_filename)
        self.api.download_from_url(job_url, target_result_zip)
        zip_file.unzip_file(target_result_zip, out_dir=self.context.local_root)
        try:
            os.makedirs(os.path.join(self.context.local_root, "backup"), exist_ok=True)
            shutil.move(
                target_result_zip,
                os.path.join(
                    self.context.local_root,
                    "backup",
                    os.path.split(target_result_zip)[1],
                ),
            )
        except (OSError, shutil.Error) as e:
            dlog.exception("unable to backup file, " + str(e))

    def check_finish_tag(self, job):
        job_tag_finished = job.job_hash + "_job_tag_finished"
        dlog.info("check if job finished: ", job.job_id, job_tag_finished)
        return self.context.check_file_exists(job_tag_finished)
        # return
        # pass

    def check_if_recover(self, submission):
        return False
        # pass

    @staticmethod
    def map_dp_job_state(status, exit_code, ignore_exit_code=True):
        if isinstance(status, JobStatus):
            return status
        map_dict = {
            -1: JobStatus.terminated,
            0: JobStatus.waiting,
            1: JobStatus.running,
            2: JobStatus.finished,
            3: JobStatus.waiting,
            4: JobStatus.running,
            5: JobStatus.terminated,
            6: JobStatus.running,
            9: JobStatus.waiting,
        }
        if status not in map_dict:
            dlog.error(f"unknown job status {status}")
            return JobStatus.unknown
        if status == -1 and exit_code != 0 and ignore_exit_code:
            return JobStatus.finished
        return map_dict[status]

    def kill(self, job):
        """Kill the job.

        Parameters
        ----------
        job : Job
            job
        """
        job_id = job.job_id
        self.api.kill(job_id)

    def get_exit_code(self, job) -> int:
        job_id = self._parse_job_id(job.job_id)
        if job_id <= 0:
            raise RuntimeError(f"cannot parse job id {job.job_id}")

        check_return = self._get_job_detail(job_id, self.group_id)
        return check_return.get("exitCode", -999)  # type: ignore

    def _parse_job_id(self, str_job_id: str) -> int:
        job_id = 0
        if "job_group_id" in str_job_id:
            ids = str_job_id.split(":job_group_id:")
            job_id, _ = int(ids[0]), int(ids[1])
        return job_id


DpCloudServer = Bohrium
Lebesgue = Bohrium



import os
import sys
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
from .context import (
    dpdispatcher,
    setUpModule,  # noqa: F401
)


class TestImportClasses(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_import_class_Machine(self):
        from dpdispatcher import Machine

        self.assertEqual(dpdispatcher.machine.Machine, Machine)

    def test_import_class_Resources(self):
        from dpdispatcher import Resources

        self.assertEqual(dpdispatcher.submission.Resources, Resources)

    def test_import_class_Submission(self):
        from dpdispatcher import Submission

        self.assertEqual(dpdispatcher.submission.Submission, Submission)

    def test_import_class_Task(self):
        from dpdispatcher import Task

        self.assertEqual(dpdispatcher.submission.Task, Task)



import os
import shutil
import sys
import unittest
from unittest.mock import MagicMock

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
from .context import (
    LazyLocalContext,
    setUpModule,  # noqa: F401
)


class TestLazyLocalContext(unittest.TestCase):
    def setUp(self):
        # os.makedirs('loc', exist_ok = True)
        # os.makedirs('loc/task0', exist_ok = True)
        # os.makedirs('loc/task1', exist_ok = True)
        shutil.copytree(src="test_context_dir/", dst="tmp_lazy_local_context_dir/")

        self.lazy_local_context = LazyLocalContext(
            local_root="tmp_lazy_local_context_dir/"
        )
        submission = MagicMock(work_base="0_md/")
        self.lazy_local_context.bind_submission(submission)

    def tearDown(self):
        shutil.rmtree("tmp_lazy_local_context_dir/")

    def test_upload(self):
        pass

    def test_download(self):
        pass

    # TODO: support other platforms
    @unittest.skipIf(sys.platform != "linux", "not linux")
    def test_block_call(self):
        code, stdin, stdout, stderr = self.lazy_local_context.block_call("ls")
        self.assertEqual(
            stdout.readlines(),
            [
                "bct-1\n",
                "bct-2\n",
                "bct-3\n",
                "bct-4\n",
                "dir with space\n",
                "graph.pb\n",
                "some_dir\n",
            ],
        )
        self.assertEqual(code, 0)

        code, stdin, stdout, stderr = self.lazy_local_context.block_call("ls a")
        self.assertEqual(code, 2)
        # self.assertEqual(stderr.read().decode('utf-8'), "ls: cannot access 'a': No such file or directory\n")
        err_msg = stderr.read().decode("utf-8")
        self.assertTrue("ls: cannot access" in err_msg)
        self.assertTrue("No such file or directory\n" in err_msg)

    # def test_block_checkcall(self) :
    #     self.job  = LazyLocalContext('loc', None)
    #     tasks = ['task0', 'task1']
    #     files = ['test0', 'test1']
    #     self.job.upload(tasks, files)
    #     # ls
    #     stdin, stdout, stderr = self.job.block_checkcall('ls')
    #     self.assertEqual(stdout.read().decode('utf-8'), 'task0\ntask1\n')
    #     self.assertEqual(stdout.readlines(), ['task0\n','task1\n'])
    #     with self.assertRaises(RuntimeError):
    #         stdin, stdout, stderr = self.job.block_checkcall('ls a')

    # def test_file(self) :
    #     self.job = LazyLocalContext('loc', None)
    #     self.assertFalse(self.job.check_file_exists('aaa'))
    #     tmp = str(uuid.uuid4())
    #     self.job.write_file('aaa', tmp)
    #     self.assertTrue(self.job.check_file_exists('aaa'))
    #     tmp1 = self.job.read_file('aaa')
    #     self.assertEqual(tmp, tmp1)

    # def test_call(self) :
    #     self.job = LazyLocalContext('loc', None)
    #     proc = self.job.call('sleep 3')
    #     self.assertFalse(self.job.check_finish(proc))
    #     time.sleep(1)
    #     self.assertFalse(self.job.check_finish(proc))
    #     time.sleep(2.5)
    #     self.assertTrue(self.job.check_finish(proc))
    #     r,o,e=self.job.get_return(proc)
    #     self.assertEqual(r, 0)
    #     self.assertEqual(o.read(), b'')
    #     self.assertEqual(e.read(), b'')
    #     r,o,e=self.job.get_return(proc)
    #     self.assertEqual(r, 0)
    #     self.assertEqual(o, None)
    #     self.assertEqual(e, None)



import os
import sys
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"

from .context import Machine, Resources, Task, setUpModule  # noqa: F401


class TestJob(unittest.TestCase):
    def test_machine_argcheck(self):
        norm_dict = Machine.load_from_dict(
            {
                "batch_type": "slurm",
                "context_type": "local",
                "local_root": "./",
                "remote_root": "/some/path",
            }
        ).serialize()
        expected_dict = {
            "batch_type": "Slurm",
            "context_type": "LocalContext",
            "local_root": "./",
            "remote_root": "/some/path",
            "remote_profile": {},
        }
        self.assertDictEqual(norm_dict, expected_dict)

    def test_resources_argcheck(self):
        norm_dict = Resources.load_from_dict(
            {
                "number_node": 1,
                "cpu_per_node": 2,
                "gpu_per_node": 0,
                "queue_name": "haha",
                "group_size": 1,
                "envs": {
                    "aa": "bb",
                },
                "kwargs": {
                    "cc": True,
                },
            }
        ).serialize()
        expected_dict = {
            "append_script": [],
            "cpu_per_node": 2,
            "custom_flags": [],
            "envs": {"aa": "bb"},
            "gpu_per_node": 0,
            "group_size": 1,
            "kwargs": {
                "cc": True,
            },
            "module_list": [],
            "module_purge": False,
            "module_unload_list": [],
            "number_node": 1,
            "para_deg": 1,
            "prepend_script": [],
            "queue_name": "haha",
            "source_list": [],
            "strategy": {"if_cuda_multi_devices": False, "ratio_unfinished": 0.0},
            "wait_time": 0,
        }
        self.assertDictEqual(norm_dict, expected_dict)

    def test_task_argcheck(self):
        norm_dict = Task.load_from_dict(
            {
                "command": "ls",
                "task_work_path": "./",
                "forward_files": [],
                "backward_files": [],
                "outlog": "out",
                "errlog": "err",
            }
        ).serialize()
        expected_dict = {
            "command": "ls",
            "task_work_path": "./",
            "forward_files": [],
            "backward_files": [],
            "outlog": "out",
            "errlog": "err",
        }
        self.assertDictEqual(norm_dict, expected_dict)



import os
import sys
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
# from .context import FinRecord
from .context import _split_tasks

# class TestFinRecord(unittest.TestCase):
#     def setUp(self):
#         self.njobs = 10
#         self.fr = FinRecord('.', self.njobs)

#     def tearDown(self):
#         if os.path.isfile('fin.record'):
#             os.remove('fin.record')

#     def test_all_false(self) :
#         recd = self.fr.get_record()
#         self.assertEqual(recd, [False]*self.njobs)

#     def test_write_read(self) :
#         recd = self.fr.get_record()
#         recd[self.njobs//3] = True
#         self.fr.write_record(recd)
#         recd1 = self.fr.get_record()
#         self.assertEqual(recd, recd1)


class TestDispatchSplit(unittest.TestCase):
    def test_split(self):
        tasks = [ii for ii in range(10)]
        chunks = _split_tasks(tasks, 5)
        self.assertEqual(chunks, [[0, 2, 4, 6, 8], [1, 3, 5, 7, 9]])
        # self.assertEqual(chunks, [[0,2,4,6,10],[1,3,5,7,9]])

    def test_split_1(self):
        tasks = [ii for ii in range(13)]
        chunks = _split_tasks(tasks, 5)
        self.assertEqual(chunks, [[0, 3, 6, 9, 12], [1, 4, 7, 10], [2, 5, 8, 11]])



import getpass
import os
import shutil
import sys
import unittest
import uuid

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
from .context import SSHContext, SSHSession


class TestSSHContext(unittest.TestCase):
    def setUp(self):
        os.makedirs("loc", exist_ok=True)
        os.makedirs("loc/task0", exist_ok=True)
        os.makedirs("loc/task1", exist_ok=True)
        for ii in ["loc/task0", "loc/task1"]:
            with open(os.path.join(ii, "test0"), "w") as fp:
                fp.write(str(uuid.uuid4()))
            with open(os.path.join(ii, "test1"), "w") as fp:
                fp.write(str(uuid.uuid4()))
            os.makedirs(os.path.join(ii, "dir0"), exist_ok=True)
            with open(os.path.join(ii, "dir0", "test2"), "w") as fp:
                fp.write(str(uuid.uuid4()))
        os.makedirs("rmt", exist_ok=True)
        try:
            self.ssh_session = SSHSession(
                {
                    "hostname": "localhost",
                    "port": 22,
                    "username": getpass.getuser(),
                    "work_path": os.path.join(os.getcwd(), "rmt"),
                }
            )
        except Exception:
            # for tianhe-2
            try:
                self.ssh_session = SSHSession(
                    {
                        "hostname": "localhost",
                        "port": 5566,
                        "username": getpass.getuser(),
                        "work_path": os.path.join(os.getcwd(), "rmt"),
                    }
                )
            except Exception:
                self.skipTest("Network error")
        self.job = SSHContext("loc", self.ssh_session)
        self.job1 = SSHContext("loc", self.ssh_session, job_uuid=self.job.job_uuid)

    def tearDown(self):
        self.job.close()
        self.job1.close()
        shutil.rmtree("loc")
        shutil.rmtree("rmt")

    def test_upload(self):
        tasks = ["task0", "task1"]
        files = ["test0", "test1"]
        self.job.upload(tasks, files)
        for ii in tasks:
            for jj in files:
                locf = os.path.join("loc", ii, jj)
                rmtf = os.path.join("rmt", self.job.job_uuid, ii, jj)
                with open(locf) as fp:
                    locs = fp.read()
                with open(rmtf) as fp:
                    rmts = fp.read()
                self.assertEqual(locs, rmts)
        self.job.upload(tasks, ["dir0"])
        for ii in tasks:
            for jj in ["dir0"]:
                for kk in ["test2"]:
                    locf = os.path.join("loc", ii, jj, kk)
                    rmtf = os.path.join("rmt", self.job.job_uuid, ii, jj, kk)
                    with open(locf) as fp:
                        locs = fp.read()
                    with open(rmtf) as fp:
                        rmts = fp.read()
                    self.assertEqual(locs, rmts)

    def test_donwload(self):
        tasks = ["task0", "task1"]
        self.job.upload(tasks, ["test0", "dir0"])
        # generate extra donwload files
        record_uuid = []
        for ii in tasks:
            for jj in ["test4", "test5"]:
                with open(os.path.join("rmt", self.job.job_uuid, ii, jj), "w") as fp:
                    tmp = str(uuid.uuid4())
                    fp.write(tmp)
                    record_uuid.append(tmp)
        # generate extra donwload dirs and files
        for ii in tasks:
            for jj in ["dir1"]:
                os.makedirs(
                    os.path.join("rmt", self.job.job_uuid, ii, jj), exist_ok=False
                )
                for kk in ["test6"]:
                    with open(
                        os.path.join("rmt", self.job.job_uuid, ii, jj, kk), "w"
                    ) as fp:
                        tmp = str(uuid.uuid4())
                        fp.write(tmp)
                        record_uuid.append(tmp)
        # donwload
        files = ["test4", "test5", "dir1"]
        self.job.download(tasks, files)
        # check dlded
        cc = 0
        for ii in tasks:
            for jj in ["test4", "test5"]:
                with open(os.path.join("loc", ii, jj)) as fp:
                    tmp = fp.read()
                    self.assertEqual(tmp, record_uuid[cc])
                    cc += 1
        for ii in tasks:
            for jj in ["dir1"]:
                for kk in ["test6"]:
                    with open(os.path.join("loc", ii, jj, kk)) as fp:
                        tmp = fp.read()
                        self.assertEqual(tmp, record_uuid[cc])
                        cc += 1

    def test_donwload_check_mark(self):
        tasks = ["task0", "task1"]
        self.job.upload(tasks, ["test0", "dir0"])
        # generate extra donwload files
        record_uuid = []
        for ii in tasks:
            for jj in ["test6", "test7"]:
                if (ii == "task1" and jj == "test7") or (
                    ii == "task0" and jj == "test6"
                ):
                    continue
                with open(os.path.join("rmt", self.job.job_uuid, ii, jj), "w") as fp:
                    tmp = str(uuid.uuid4())
                    fp.write(tmp)
                    record_uuid.append(tmp)
        # donwload
        files = ["test6", "test7", "dir1"]
        self.job.download(tasks, files, check_exists=True, mark_failure=True)
        # check dlded
        cc = 0
        for ii in tasks:
            for jj in ["test6", "test7"]:
                if (ii == "task1" and jj == "test7") or (
                    ii == "task0" and jj == "test6"
                ):
                    self.assertFalse(
                        os.path.exists(os.path.join("loc", ii, jj)),
                        msg="found " + os.path.join("loc", ii, jj),
                    )
                    self.assertTrue(
                        os.path.exists(
                            os.path.join("loc", ii, "tag_failure_download_%s" % jj)
                        ),
                        msg="failed to find "
                        + os.path.join("loc", ii, "tag_failure_download_%s" % jj),
                    )
                    continue
                with open(os.path.join("loc", ii, jj)) as fp:
                    tmp = fp.read()
                    self.assertEqual(tmp, record_uuid[cc])
                    cc += 1
        for ii in tasks:
            for jj in ["dir1"]:
                self.assertFalse(os.path.exists(os.path.join("loc", ii, jj)))
                self.assertTrue(
                    os.path.exists(
                        os.path.join("loc", ii, "tag_failure_download_%s" % jj)
                    )
                )

    def test_donwload_check_nomark(self):
        tasks = ["task0", "task1"]
        self.job.upload(tasks, ["test0", "dir0"])
        # generate extra donwload files
        record_uuid = []
        for ii in tasks:
            for jj in ["test6", "test7"]:
                if ii == "task1" and jj == "test7":
                    continue
                if ii == "task0" and jj == "test6":
                    continue
                with open(os.path.join("rmt", self.job.job_uuid, ii, jj), "w") as fp:
                    tmp = str(uuid.uuid4())
                    fp.write(tmp)
                    record_uuid.append(tmp)
        # donwload
        files = ["test6", "test7", "dir1"]
        self.job.download(tasks, files, check_exists=True, mark_failure=False)
        # check dlded
        cc = 0
        for ii in tasks:
            for jj in ["test6", "test7"]:
                if ii == "task1" and jj == "test7":
                    self.assertFalse(
                        os.path.exists(os.path.join("loc", ii, jj)),
                        msg="found " + os.path.join("loc", ii, jj),
                    )
                    self.assertFalse(
                        os.path.exists(
                            os.path.join("loc", ii, "tag_failure_download_%s" % jj)
                        ),
                        msg="found "
                        + os.path.join("loc", ii, "tag_failure_download_%s" % jj),
                    )
                    continue
                if ii == "task0" and jj == "test6":
                    self.assertFalse(
                        os.path.exists(os.path.join("loc", ii, jj)),
                        msg="found " + os.path.join("loc", ii, jj),
                    )
                    self.assertFalse(
                        os.path.exists(
                            os.path.join("loc", ii, "tag_failure_download_%s" % jj)
                        ),
                        msg="found "
                        + os.path.join("loc", ii, "tag_failure_download_%s" % jj),
                    )
                    continue
                with open(os.path.join("loc", ii, jj)) as fp:
                    tmp = fp.read()
                    self.assertEqual(tmp, record_uuid[cc])
                    cc += 1
        for ii in tasks:
            for jj in ["dir1"]:
                self.assertFalse(os.path.exists(os.path.join("loc", ii, jj)))
                self.assertFalse(
                    os.path.exists(
                        os.path.join("loc", ii, "tag_failure_download_%s" % jj)
                    )
                )

    def test_block_call(self):
        tasks = ["task0", "task1"]
        files = ["test0", "test1"]
        self.job.upload(tasks, files)
        # ls
        code, stdin, stdout, stderr = self.job.block_call("ls")
        self.assertEqual(stdout.read(), b"task0\ntask1\n")
        self.assertEqual(code, 0)
        code, stdin, stdout, stderr = self.job.block_call("ls")
        self.assertEqual(stdout.readlines(), ["task0\n", "task1\n"])
        code, stdin, stdout, stderr = self.job.block_call("ls a")
        self.assertEqual(code, 2)
        # self.assertEqual(stderr.read().decode('utf-8'), 'ls: cannot access a: No such file or directory\n')
        err_msg = stderr.read().decode("utf-8")
        self.assertTrue("ls: cannot access" in err_msg)
        self.assertTrue("No such file or directory\n" in err_msg)

    def test_block_checkcall(self):
        tasks = ["task0", "task1"]
        files = ["test0", "test1"]
        self.job.upload(tasks, files)
        # ls
        stdin, stdout, stderr = self.job.block_checkcall("ls")
        self.assertEqual(stdout.read(), b"task0\ntask1\n")
        stdin, stdout, stderr = self.job.block_checkcall("ls")
        self.assertEqual(stdout.readlines(), ["task0\n", "task1\n"])
        with self.assertRaises(RuntimeError):
            stdin, stdout, stderr = self.job.block_checkcall("ls a")

    def test_file(self):
        self.assertFalse(self.job.check_file_exists("aaa"))
        tmp = str(uuid.uuid4())
        self.job.write_file("aaa", tmp)
        self.assertTrue(self.job.check_file_exists("aaa"))
        tmp1 = self.job.read_file("aaa")
        self.assertEqual(tmp, tmp1)



import subprocess as sp
import unittest


class TestCLI(unittest.TestCase):
    def test_cli(self):
        sp.check_output(["dpdisp", "-h"])
        for subcommand in ("gui",):
            sp.check_output(["dpdisp", subcommand, "-h"])



# from dpdispatcher.batch_object import BatchObject
# from dpdispatcher.batch import Batch
import os
import sys
import tempfile
import textwrap

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
import json
import unittest

from .context import (
    Machine,
    Resources,
    Submission,
    Task,
    setUpModule,  # noqa: F401
)


class TestLSFScriptGeneration(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def test_shell_trival(self):
        with open("jsons/machine_lazy_local_lsf.json") as f:
            machine_dict = json.load(f)

        machine = Machine(**machine_dict["machine"])
        resources = Resources(**machine_dict["resources"])

        task1 = Task(
            command="cat example.txt",
            task_work_path="dir1/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task2 = Task(
            command="cat example.txt",
            task_work_path="dir2/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task3 = Task(
            command="cat example.txt",
            task_work_path="dir3/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task4 = Task(
            command="cat example.txt",
            task_work_path="dir4/",
            forward_files=["example.txt"],
            backward_files=["out.txt"],
            outlog="out.txt",
        )
        task_list = [task1, task2, task3, task4]

        submission = Submission(
            work_base="parent_dir/",
            machine=machine,
            resources=resources,
            forward_common_files=["graph.pb"],
            backward_common_files=[],
            task_list=task_list,
        )
        submission.generate_jobs()

        task_hash = submission.get_hash()
        test_job = submission.belonging_jobs[0]
        job_hash = test_job.job_hash

        header_str = machine.gen_script_header(test_job)
        benchmark_header = textwrap.dedent(
            """\
            #!/bin/bash -l
            #BSUB -e %J.err
            #BSUB -o %J.out
            #BSUB -n 4
            #BSUB -R 'span[ptile=4]'
            #BSUB -q gpu
            #BSUB -gpu 'num=1:mode=shared:j_exclusive=no'"""
        )
        self.assertEqual(header_str, benchmark_header)

        custom_flags_str = machine.gen_script_custom_flags_lines(test_job)
        benchmark_custom_flags = textwrap.dedent(
            """\
            #BSUB -R "select[hname != g005]"
            #BSUB -W 24:00
            """
        )
        self.assertEqual(custom_flags_str, benchmark_custom_flags)

        env_str = machine.gen_script_env(test_job)
        benchmark_env = textwrap.dedent(
            f"""
            REMOTE_ROOT={task_hash}
            echo 0 > $REMOTE_ROOT/{job_hash}_flag_if_job_task_fail
            test $? -ne 0 && exit 1

            module purge

            module load use.own
            module load deepmd/1.3

            {{ source /data/home/ypliu/scripts/avail_gpu.sh; }} 
            {{ source /data/home/ypliu/dprun/tf_envs.sh; }} 

            export DP_DISPATCHER_EXPORT=test_foo_bar_baz

            echo 'The summer you were there.'
            """
        )
        self.assertEqual(env_str.split("\n")[2:], benchmark_env.split("\n")[2:])

        footer_str = machine.gen_script_end(test_job)
        benchmark_footer = textwrap.dedent(
            f"""\

            cd $REMOTE_ROOT
            test $? -ne 0 && exit 1

            wait
            FLAG_IF_JOB_TASK_FAIL=$(cat {job_hash}_flag_if_job_task_fail)
            if test $FLAG_IF_JOB_TASK_FAIL -eq 0; then touch {job_hash}_job_tag_finished; else exit 1;fi

            echo 'shizuku'
            echo 'kaori'
            """
        )
        self.assertEqual(footer_str, benchmark_footer)

    @unittest.skipIf(sys.platform == "win32", "skip for persimission error")
    def test_template(self):
        with open("jsons/machine_lazy_local_lsf.json") as f:
            machine_dict = json.load(f)

        benchmark_str = textwrap.dedent(
            """\
            #!/bin/bash -l
            #BSUB -e %J.err
            #BSUB -o %J.out
            #BSUB -n 4
            #BSUB -R 'span[ptile=4]'
            #BSUB -q gpu
            #BSUB -gpu 'num=1:mode=shared:j_exclusive=no'"""
        )

        with tempfile.NamedTemporaryFile("w") as f:
            f.write(benchmark_str)
            f.flush()

            machine_dict["resources"]["strategy"][
                "customized_script_header_template_file"
            ] = f.name

            machine = Machine(**machine_dict["machine"])
            resources = Resources(**machine_dict["resources"])

            task1 = Task(
                command="cat example.txt",
                task_work_path="dir1/",
                forward_files=["example.txt"],
                backward_files=["out.txt"],
                outlog="out.txt",
            )
            task2 = Task(
                command="cat example.txt",
                task_work_path="dir2/",
                forward_files=["example.txt"],
                backward_files=["out.txt"],
                outlog="out.txt",
            )
            task3 = Task(
                command="cat example.txt",
                task_work_path="dir3/",
                forward_files=["example.txt"],
                backward_files=["out.txt"],
                outlog="out.txt",
            )
            task4 = Task(
                command="cat example.txt",
                task_work_path="dir4/",
                forward_files=["example.txt"],
                backward_files=["out.txt"],
                outlog="out.txt",
            )
            task_list = [task1, task2, task3, task4]

            submission = Submission(
                work_base="parent_dir/",
                machine=machine,
                resources=resources,
                forward_common_files=["graph.pb"],
                backward_common_files=[],
                task_list=task_list,
            )
            submission.generate_jobs()
            str = machine.gen_script_header(submission.belonging_jobs[0])
            self.assertEqual(str, benchmark_str)



import os
import sys

sys.path.insert(
    0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", ".."))
)

from dpgen.dispatcher.Dispatcher import Dispatcher  # noqa: F401
from dpgen.dispatcher.JobStatus import JobStatus  # noqa: F401
from dpgen.dispatcher.LocalContext import LocalContext, LocalSession  # noqa: F401
from dpgen.dispatcher.PBS import PBS  # noqa: F401
from dpgen.dispatcher.SSHContext import SSHContext, SSHSession  # noqa: F401


def my_file_cmp(test, f0, f1):
    with open(f0) as fp0:
        with open(f1) as fp1:
            test.assertTrue(fp0.read() == fp1.read())


def setUpModule():
    os.chdir(os.path.abspath(os.path.dirname(__file__)))



# /usr/bin/python

import os

from dpdispatcher.utils import run_cmd_with_all_output


class HDFS:
    """Fundamental class for HDFS basic manipulation."""

    @staticmethod
    def exists(uri):
        """Check existence of hdfs uri
        Returns: True on exists
        Raises: RuntimeError.
        """
        cmd = f"hadoop fs -test -e {uri}"
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True
            elif ret == 1:
                return False
            else:
                raise RuntimeError(
                    f"Cannot check existence of hdfs uri[{uri}] "
                    f"with cmd[{cmd}]; ret[{ret}] stdout[{out}] stderr[{err}]"
                )
        except Exception as e:
            raise RuntimeError(
                f"Cannot check existence of hdfs uri[{uri}] " f"with cmd[{cmd}]"
            ) from e

    @staticmethod
    def remove(uri):
        """Check existence of hdfs uri
        Returns: True on exists
        Raises: RuntimeError.
        """
        cmd = f"hadoop fs -rm -r {uri}"
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True
            else:
                raise RuntimeError(
                    f"Cannot remove hdfs uri[{uri}] "
                    f"with cmd[{cmd}]; ret[{ret}] output[{out}] stderr[{err}]"
                )
        except Exception as e:
            raise RuntimeError(
                f"Cannot remove hdfs uri[{uri}] " f"with cmd[{cmd}]"
            ) from e

    @staticmethod
    def mkdir(uri):
        """Make new hdfs directory
        Returns: True on success
        Raises: RuntimeError.
        """
        cmd = f"hadoop fs -mkdir -p {uri}"
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True
            else:
                raise RuntimeError(
                    f"Cannot mkdir of hdfs uri[{uri}] "
                    f"with cmd[{cmd}]; ret[{ret}] output[{out}] stderr[{err}]"
                )
        except Exception as e:
            raise RuntimeError(
                f"Cannot mkdir of hdfs uri[{uri}] " f"with cmd[{cmd}]"
            ) from e

    @staticmethod
    def copy_from_local(local_path, to_uri):
        """Returns: True on success
        Raises: on unexpected error.
        """
        # Make sure local_path is accessible
        if not os.path.exists(local_path) or not os.access(local_path, os.R_OK):
            raise RuntimeError(f"try to access local_path[{local_path}] " "but failed")
        cmd = f"hadoop fs -copyFromLocal -f {local_path} {to_uri}"
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True, out
            else:
                raise RuntimeError(
                    "Cannot copy local[{}] to remote[{}] with cmd[{}]; "
                    "ret[{}] output[{}] stderr[{}]".format(
                        local_path, to_uri, cmd, ret, out, err
                    )
                )
        except Exception as e:
            raise RuntimeError(
                f"Cannot copy local[{local_path}] to remote[{to_uri}] with cmd[{cmd}]"
            ) from e

    @staticmethod
    def copy_to_local(from_uri, local_path):
        remote = ""
        if isinstance(from_uri, str):
            remote = from_uri
        elif isinstance(from_uri, list) or isinstance(from_uri, tuple):
            remote = " ".join(from_uri)
        cmd = f"hadoop fs -copyToLocal {remote} {local_path}"

        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True
            else:
                raise RuntimeError(
                    "Cannot copy remote[{}] to local[{}] with cmd[{}]; "
                    "ret[{}] output[{}] stderr[{}]".format(
                        from_uri, local_path, cmd, ret, out, err
                    )
                )
        except Exception as e:
            raise RuntimeError(
                f"Cannot copy remote[{from_uri}] to local[{local_path}] with cmd[{cmd}]"
            ) from e

    @staticmethod
    def read_hdfs_file(uri):
        cmd = f"hadoop fs -text {uri}"
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return out
            else:
                raise RuntimeError(
                    f"Cannot read text from uri[{uri}]"
                    f"cmd [{cmd}] ret[{ret}] output[{out}] stderr[{err}]"
                )
        except Exception as e:
            raise RuntimeError(
                f"Cannot read text from uri[{uri}]" f"cmd [{cmd}]"
            ) from e

    @staticmethod
    def move(from_uri, to_uri):
        cmd = f"hadoop fs -mv {from_uri} {to_uri}"
        try:
            ret, out, err = run_cmd_with_all_output(cmd)
            if ret == 0:
                return True
            else:
                raise RuntimeError(
                    f"Cannot move from_uri[{from_uri}] to "
                    f"to_uri[{to_uri}] with cmd[{cmd}]; "
                    f"ret[{ret}] output[{out}] stderr[{err}]"
                )
        except Exception as e:
            raise RuntimeError(
                f"Cannot move from_uri[{from_uri}] to "
                f"to_uri[{to_uri}] with cmd[{cmd}]"
            ) from e



import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from dpdispatcher.lazy_local_context import LazyLocalContext
from dpdispatcher.pbs import PBS
from dpdispatcher.submission import Resources, Submission, Task

# local_session = LocalSession({'work_path':'temp2'})
# local_context = LocalContext(local_root='temp1/0_md', work_profile=local_session)
lazy_local_context = LazyLocalContext(
    local_root="/home/fengbo/10_dpdispatcher/dpdispatcher/tests/temp3",
    work_profile=None,
)

# pbs = PBS(context=local_context)
pbs = PBS(context=lazy_local_context)

resources = Resources(
    number_node=1,
    cpu_per_node=4,
    gpu_per_node=1,
    queue_name="V100_8_32",
    group_size=4,
    if_cuda_multi_devices=True,
)
submission = Submission(work_base="0_md", resources=resources)
task1 = Task(command="lmp_serial -i input.lammps", task_work_path="bct-1")
task2 = Task(command="lmp_serial -i input.lammps", task_work_path="bct-2")
task3 = Task(command="lmp_serial -i input.lammps", task_work_path="bct-3")
task4 = Task(command="lmp_serial -i input.lammps", task_work_path="bct-4")
submission.register_task_list(
    [
        task1,
        task2,
        task3,
        task4,
    ]
)
submission.generate_jobs()
submission.bind_batch(batch=pbs)
# for job in submission.belonging_jobs:
#     job.job_to_json()
# print('111', submission)
# submission2 = Submission.recover_jobs_from_json('./jr.json')
# print('222', submission2)
# print(submission==submission2)
submission.run_submission()

# submission1.dump_jobs_fo_json()
# submission2 = Submission.submission_from_json('jsons/submission.json')
# print(677, submission==submission2)
# print(submission1.belonging_jobs)
# print(local_context)



import os
import sys
import unittest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
__package__ = "tests"
from .context import LocalSession


class TestLocalSession(unittest.TestCase):
    def test_work_path(self):
        cwd = os.getcwd()
        wp = LocalSession({"work_path": cwd})
        self.assertTrue(os.path.abspath(cwd), wp.get_work_root())
